;;;=======================================================
;;;  Assignment A2 - CIS666 Expert Systems
;;;  Author: Jungkee Kim
;;;
;;;  Decision Making of the Admission Program
;;;  Input:
;;;    No terminal input but collected data for applicants
;;;  Output:
;;;    the list of granted applicants and rejected
;;;    applicants
;;;=======================================================

; **************
; FACT TEMPLATES
; **************

(deftemplate applicant-record
  (slot name
    (type STRING))
  (slot id
    (type INTEGER))
  (slot cs
    (type SYMBOL))
  (slot gpa
    (type FLOAT))
  (slot adjusted-gpa		;; initially 0.0
    (type FLOAT))
  (multislot gre
    (type INTEGER))
  (slot toefl
    (type INTEGER))
  (slot consider-personal-history
    (type SYMBOL)
    (allowed-symbols y n))
  (slot cs-job-experience
    (type SYMBOL)
    (allowed-symbols y n))
  (slot exceptional-research
    (type SYMBOL)
    (allowed-symbols y n))
  (slot outstanding-school
    (type SYMBOL)
    (allowed-symbols y n))
  (slot got-cs-courses
    (type SYMBOL)
    (allowed-symbols y n))
  (slot low-eng
    (type SYMBOL)
    (allowed-symbols y n))
)

; **********
; INITIALIZE
; **********

(defrule initialize
  ?t <- (initial-fact)
  (not (applicant-record (name ?)))
  =>
  (load-facts "Admission.dat")
  (retract ?t)
)

; *********
; FUNCTIONS
; *********

(deffunction assert-grant (?name ?id)
  (assert (granted ?name ?id))
  ;(printout t ?name "(" ?id ")" " is granted." crlf)
  (return TRUE)
)

(deffunction print-reject (?name ?id)
  (printout t ?name "(" ?id ")" " is rejected." crlf)
  (return TRUE)
)

; *****
; RULES 
; *****

(defrule adjust-gpa
  ?t <- (applicant-record (gpa ?g) (adjusted-gpa 0.0) (outstanding-school ?os))
  (applicant-record (gpa ?g) (adjusted-gpa 0.0) (consider-personal-history ?cph))
  =>
  (if (or (eq ?os y) (eq ?cph y))
    then
    (modify ?t (adjusted-gpa (+ ?g 0.2)))
    else
    (modify ?t (adjusted-gpa ?g))
  )
)

(defrule check-minimum-quantitative
  ?rec <- (applicant-record (name ?n) (id ?i) (gre ?v ?q ?a))
  =>
  (if (< ?q 700)
    then
    (print-reject ?n ?i)
    (retract ?rec)
    else
    (assert (quantitative-checked ?i))
  )
)

(defrule minimum-toefl
  ?rec <- (applicant-record (name ?n) (id ?i) (toefl ?t) (low-eng ?lg))
  =>
  (if (and (< ?t 500) (not (= ?t 0)))  ;;If TOEFL score is zero, no toefl taken.
    then
    (print-reject ?n ?i)
    (retract ?rec)
    else
    (assert (toefl-checked ?i))
    (if (eq ?lg y)
      then
      (assert (eng-recommendation ?i))
    )
  )
)

(defrule admission-for-exceptional-paper
  ?rec <- (applicant-record (name ?n) (id ?i) (exceptional-research ?er))
  ?qc <- (quantitative-checked ?i)
  ?tc <- (toefl-checked ?i)
  =>
  (if (eq ?er y)
    then
    (assert-grant ?n ?i)
    (retract ?rec)
    (retract ?qc)
    (retract ?tc)
  )
  (assert (exceptional-paper-checked ?i))
)

(defrule admission-for-cs-major
  ?rec <- (applicant-record (name ?n) (id ?i) (cs ?cs) (adjusted-gpa ?ag))
  ?qc <- (quantitative-checked ?i)
  ?tc <- (toefl-checked ?i)
  ?ep <- (exceptional-paper-checked ?i)
  =>
  (if (and (eq ?cs y) (>= ?ag 3.0))
    then
    (assert-grant ?n ?i)
    (retract ?rec)
    (retract ?qc)
    (retract ?tc)
    (retract ?ep)
  )
)

(defrule admission-for-non-cs-major
  ?rec <- (applicant-record (name ?n) (id ?i) (cs ?cs) (adjusted-gpa ?ag) (got-cs-courses ?gcc))
  ?qc <- (quantitative-checked ?i)
  ?tc <- (toefl-checked ?i)
  ?ep <- (exceptional-paper-checked ?i)
  =>
  (if (and (eq ?cs n) (>= ?ag 3.2))
    then
    (assert-grant ?n ?i)
    (if (eq ?gcc n)
      then
      (assert (take-cs-course ?i))
    )
    (retract ?rec)
    (retract ?qc)
    (retract ?tc)
    (retract ?ep)
  ) 
)

(defrule admission-for-non-cs-major-with-industrial-experience
  ?rec <- (applicant-record (name ?n) (id ?i) (cs ?cs) (adjusted-gpa ?ag))
  ?qc <- (quantitative-checked ?i)
  ?tc <- (toefl-checked ?i)
  ?ep <- (exceptional-paper-checked ?i)
  =>
  (if (and (eq ?cs n) (>= ?ag 3.0))
    then
    (assert-grant ?n ?i)
    (retract ?rec)
    (retract ?qc)
    (retract ?tc)
    (retract ?ep)
  )
)

(defrule rest-out-of-granted
  ?rec <- (applicant-record (name ?n) (id ?i) (cs ?cs) (adjusted-gpa ?ag))
  ?qc <- (quantitative-checked ?i)
  ?tc <- (toefl-checked ?i)
  ?ep <- (exceptional-paper-checked ?i)
  =>
  (if (< ?ag 3.0)
    then
    (print-reject ?n ?i)
    (retract ?rec)
  )
)

(defrule print-grant-with-eng-cs
  ?g <- (granted ?n ?i)
  ?e <- (eng-recommendation ?i)
  ?t <- (take-cs-course ?i)
  =>
  (retract ?g)
  (retract ?e)
  (retract ?t)
  (printout t ?n "(" ?i ") is granted with taking ENG and CS courses." crlf)
)

(defrule print-grant-with-eng
  ?g <- (granted ?n ?i)
  ?e <- (eng-recommendation ?i)
  (not (take-cs-course ?i))
  =>
  (retract ?g)
  (retract ?e)
  (printout t ?n "(" ?i ") is granted with taking ENG courses." crlf)
)

(defrule print-grant-with-cs
  ?g <- (granted ?n ?i)
  (not (eng-recommendation ?i))
  ?t <- (take-cs-course ?i)
  =>
  (retract ?g)
  (retract ?t)
  (printout t ?n "(" ?i ") is granted with taking CS courses." crlf)
)

(defrule print-grant-without-condition
  ?g <- (granted ?n ?i)
  (not (eng-recommendation ?i))
  (not (take-cs-course ?i))
  =>
  (retract ?g)
  (printout t ?n "(" ?i ") is granted." crlf)
)
