;;;======================================================
;;;  Assignment A1 - CIS666 Expert Systems
;;;  Author: Jungkee Kim
;;;
;;;  Course Recommandation Program
;;;  Input:
;;;    user name, unique student id
;;;  Output:
;;;    the list of courses already taken and their grades
;;;    a list of courses that the student must take
;;;======================================================

; **************
; FACT TEMPLATES
; **************

(deftemplate course-history
  (slot name
    (type SYMBOL))
  (slot id
    (type INTEGER))
  (slot field
    (type SYMBOL))
  (multislot courses))

(deftemplate course-properties
  (slot name
    (type SYMBOL))
  (slot minimum-grade
    (type SYMBOL)
    (allowed-symbols B C D))
  (slot offer
    (type SYMBOL)
    (allowed-symbols Fall Spring both))
  (multislot prerequisites
    (type SYMBOL)))

;; Courses for the graduation are fixed.
(deftemplate courses-for-graduate
  (slot name
    (type SYMBOL))
  (multislot courses
    (type SYMBOL)))

; **********
; INITIALIZE
; **********

(defrule initialize
  (not (courses-for-graduate (name ?)))
  =>
  (load-facts "CourseList.dat"))

(defrule prompt
  (initial-fact)
  =>
  (printout t "Enter your name" crlf)
  (bind ?name (read))
  (printout t "Enter your ID" crlf)
  (bind ?id (read))
  (assert (input-info ?name ?id))
  (printout t "Enter the semester (Fall/Spring)" crlf)
  (bind ?semester (read))
  (assert (current-semester ?semester))
  (assert (passed-list))
  (assert (count 6))			;classes per semester
)

; *********
; FUNCTIONS
; *********

(deffunction is-passed (?grade ?mg)
  (if (eq ?grade A)
    then
    (return TRUE)
    else (if (and (eq ?grade B) (neq ?mg A))
    then
    (return TRUE)
    else (if (and (eq ?grade C) (and (neq ?mg A) (neq ?mg B)))
    then
    (return TRUE)
    else (if (and (eq ?grade D) (or (eq ?mg D) (eq ?mg F)))
    then
    (return TRUE)
    else (if (eq ?mg F)
    then
    (return TRUE)
    else
    (return FALSE)))))
  )
)

(deffunction is-prerequisite-clear (?prerequisites $?passed-list)
  ;(printout t "pre:" ?prerequisites " PL:" $?passed-list crlf)
  (if (= (length $?prerequisites) 0)
    then
    ;(printout t "length is zero" crlf)
    (return TRUE)
  )
  (if (subsetp ?prerequisites (create$ $?passed-list))
    then
    ;(printout t "all clear" crlf)
    (return TRUE)
    else
    ;(printout t "no clear" crlf)
    (return FALSE)
  )
)

; *****
; RULES 
; *****

(defrule no-name
  ?i <- (input-info ?name ?id)
  (not (course-history (name ?name)))
  =>
  (printout t "You are not on the list" crlf)
  (retract ?i)
)

(defrule no-id
  ?i <- (input-info ?name ?id)
  (not (course-history (name ?name) (id ?id)))
  =>
  (printout t "Your input for the ID is not correct" crlf)
  (retract ?i)
)

(defrule no-semester
  ?s <- (current-semester ?cs)
  =>
  (if (not (or (eq ?cs Fall) (eq ?cs Spring)))
    then
    (printout t "Semester must be Fall or Spring" crlf)
    (retract ?s)
  )
)

(defrule print-info
  ?i <- (input-info ?name ?id)
  (course-history (name ?name) (id ?id) (field ?field) (courses $?courses))
  (courses-for-graduate (name ?field) (courses $?cfg))
  =>
  (printout t "---------------" crlf)
  (printout t "Current report " crlf)
  (printout t "---------------" crlf)
  (printout t "Name: " ?name crlf "ID: " ?id crlf "Field: " ?field crlf)
  (printout t "The list of courses which you have taken:" crlf)
  (assert (authorized $?courses))
  (assert (current $?courses))
  (assert (whole-list $?cfg))
  ;(printout t "Whole-list(print-info):" $?cfg crlf)
  (retract ?i)
)

(defrule print-courses
  ?c <- (current ?course ?grade $?rest)
  =>
  (printout t ?course " : " ?grade crlf)
  (retract ?c)
  (assert (current $?rest))
)

(defrule end-of-print-courses
  ?t <- (current)
  =>
  (retract ?t)
  (printout t "---------------" crlf)
  (printout t " Next schedule " crlf)
  (printout t "---------------" crlf)
  (assert (ready-for-passed))
)

(defrule passed-course
  ?c <- (authorized ?course ?grade $?rest)
  (course-properties (name ?course) (minimum-grade ?mg))
  =>
  (if (is-passed ?grade ?mg)
    then
    (assert (passed ?course))
    else
    (assert (required ?course))
  )
  (retract ?c)
  (assert (authorized $?rest)) 
)

(defrule collect-passed
  ?p <- (passed ?course)
  ?t <- (passed-list $?pl)
  =>
  (retract ?p)
  (retract ?t)
  (assert (passed-list $?pl ?course))
  ;(printout t "collect: " $?pl ?course crlf)
)

(defrule ready-to-select
  (not (current $?courses))
  (not (passed $?pc))
  ?t <- (ready-for-passed)
  (authorized)
  =>
  ;(printout t "ready-for-slect fired" crlf)
  (retract ?t)
  (assert (ready))
)

(defrule select-required
  (ready)
  ?w <- (whole-list ?c $?rest)
  ?t <- (required ?c)
  ?u <- (count ?count)
  (course-properties (name ?c) (offer ?o))
  (current-semester ?cs)
  =>
  (retract ?w)
  (assert (whole-list $?rest))
  ;(printout t "whole-list(select-required): " $?rest crlf)
  (retract ?t)
  (if (or (eq ?o both) (eq ?o ?cs))
    then
    (printout t ?c crlf)
    (retract ?u)
    (if (> ?count 1) ;; if all six courses are selected "count" will be deleted.
      then
      (assert (count (- ?count 1)))
    )
  )
)

(defrule unselect-the-passed-course
  (ready)
  ?w <- (whole-list ?c $?rest)
  (passed-list $?pl)
  =>
  (if (member$ ?c (create$ $?pl))
    then
    (retract ?w)
    (assert (whole-list $?rest))
    ;(printout t "whole-list(unselect):" $?rest crlf)
  )
)

(defrule select-a-course
  (ready)
  ?w <- (whole-list ?c $?rest)
  ?u <- (count ?count)
  (not (required ?c))
  (course-properties (name ?c) (offer ?o) (prerequisites $?pr))
  (passed-list $?pl)
  (current-semester ?cs)
  =>
  (retract ?w)
  (assert (whole-list $?rest))
  ;(printout t ?c " whole-list(select): " $?rest crlf)
  (if (and (is-prerequisite-clear (create$ $?pr) $?pl)
           (not (member$ ?c (create$ $?pl))))
    then
    ;(printout t "cs " ?cs " o " ?o crlf)
    (if (or (eq ?o both) (eq ?o ?cs))
      then
      (printout t ?c crlf)
      (retract ?u)
      (if (> ?count 1) 
        then
        (assert (count (- ?count 1)))
      )
    )
  )
)

(defrule finalize1
  ?cs <- (current-semester ?s)
  ?at <- (authorized)
  ?pl <- (passed-list $?p)
  ?rd <- (ready)
  ?wl <- (whole-list $?w)
  (not (count $?c))
  =>
  (retract ?cs)
  (retract ?at)
  (retract ?pl)
  (retract ?rd)
  (retract ?wl)
)

(defrule finalize2
  ?cs <- (current-semester ?s)
  ?at <- (authorized)
  ?pl <- (passed-list $?p)
  ?rd <- (ready)
  ?wl <- (whole-list $?w)
  ?ct <- (count $?c)
  =>
  (retract ?cs)
  (retract ?at)
  (retract ?pl)
  (retract ?rd)
  (retract ?wl)
  (retract ?ct)
)
